package com.sugen.util;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.swing.tree.DefaultMutableTreeNode;

import com.sugen.gui.Colorable;
import com.sugen.gui.ToolTippable;
import com.sugen.util.Clade.Sequence.Annotation;
import com.sugen.util.Clade.Sequence.DomainArchitecture.Domain;

/**
 * A tree clade that adds the concepts of branch length, color,
 * geographic distribution, taxonomy, sequence, etc. 
 * Useful in clustering algorithms like phylogenetic reconstruction.
 *
 * @author Jonathan Bingham
 */
public class Clade extends DefaultMutableTreeNode 
implements Colorable, ToolTippable {
	
	public static class Property {
		private String ref;
		private String datatype;
		private String appliesTo;
		private String value;
		
		public void setRef(String ref) {
			this.ref = ref;
		}
		public String getRef() {
			return ref;
		}
		public void setDatatype(String datatype) {
			this.datatype = datatype;
		}
		public String getDatatype() {
			return datatype;
		}
		public void setAppliesTo(String appliesTo) {
			this.appliesTo = appliesTo;
		}
		public String getAppliesTo() {
			return appliesTo;
		}
		public void setValue(String value) {
			this.value = value;
		}
		public String getValue() {
			return value;
		}
	}

	public static class Confidence {
		private String type;
		private double value;

		public String getType() {
			return type;
		}
		public void setType(String type) {
			this.type = type;
		}
		public double getValue() {
			return value;
		}
		public void setValue(double value) {
			this.value = value;
		}
	}

	public static class Taxonomy {
		private String id;
		private String provider;
		private String code;
		private String scientificName;
		private String commonName;
		private String rank;
		private String uri;

		public void setScientificName(String scientificName) {
			this.scientificName = scientificName;
		}

		public String getScientificName() {
			return scientificName;
		}

		public void setCode(String code) {
			this.code = code;
		}

		public String getCode() {
			return code;
		}

		public void setCommonName(String commonName) {
			this.commonName = commonName;
		}

		public String getCommonName() {
			return commonName;
		}

		public void setRank(String rank) {
			this.rank = rank;
		}

		public String getRank() {
			return rank;
		}

		public void setUri(String uri) {
			this.uri = uri;
		}

		public String getUri() {
			return uri;
		}

		public void setId(String id) {
			this.id = id;
		}

		public String getId() {
			return id;
		}

		public void setProvider(String provider) {
			this.provider = provider;
		}

		public String getProvider() {
			return provider;
		}
	}
	
	public static class Distribution {
		public static class Point {
			private String geodeticDatum;
			private String altitudeUnit;
			private float latitude;
			private float longitude;
			private float altitude;
			
			public String getGeodeticDatum() {
				return geodeticDatum;
			}
			public void setGeodeticDatum(String geodeticDatum) {
				this.geodeticDatum = geodeticDatum;
			}
			public String getAltitudeUnit() {
				return altitudeUnit;
			}
			public void setAltitudeUnit(String altitudeUnit) {
				this.altitudeUnit = altitudeUnit;
			}
			public float getLatitude() {
				return latitude;
			}
			public void setLatitude(float latitude) {
				this.latitude = latitude;
			}
			public float getLongitude() {
				return longitude;
			}
			public void setLongitude(float longitude) {
				this.longitude = longitude;
			}
			public float getAltitude() {
				return altitude;
			}
			public void setAltitude(float altitude) {
				this.altitude = altitude;
			}
		}
		
		public static class Polygon {
			private List<Point> points;

			public List<Point> getPoints() {
				if (points == null)
					points = new ArrayList<Point>();
				return points;
			}

			public void setPoints(List<Point> points) {
				this.points = points;
			}			
		}

		private String description;
		private List<Point> points;
		private List<Polygon> polygons;

		public void setDescription(String description) {
			this.description = description;
		}

		public String getDescription() {
			return description;
		}

		public void setPoints(List<Point> points) {
			this.points = points;
		}

		public List<Point> getPoints() {
			if (points == null)
				points = new ArrayList<Point>();
			return points;
		}

		public void setPolygons(List<Polygon> polygons) {
			this.polygons = polygons;
		}

		public List<Polygon> getPolygons() {
			if (polygons == null)
				polygons = new ArrayList<Polygon>();
			return polygons;
		}
	}
	
	public static class Sequence {
		public static class Annotation {
			private String reference;
			private String source;
			private String evidence;
			private String type;
			private String description;
			private List<Confidence> confidences;
			private List<Property> properties;
			
			public void setSource(String source) {
				this.source = source;
			}
			public String getSource() {
				return source;
			}
			public void setReference(String reference) {
				this.reference = reference;
			}
			public String getReference() {
				return reference;
			}
			public void setEvidence(String evidence) {
				this.evidence = evidence;
			}
			public String getEvidence() {
				return evidence;
			}
			public void setType(String type) {
				this.type = type;
			}
			public String getType() {
				return type;
			}
			public void setDescription(String description) {
				this.description = description;
			}
			public String getDescription() {
				return description;
			}
			public List<Confidence> getConfidences() {
				if (confidences == null)
					confidences = new ArrayList<Confidence>();
				return confidences;
			}
			public List<Property> getProperties() {
				if (properties == null)
					properties = new ArrayList<Property>();
				return properties;
			}
		}
		public static class DomainArchitecture {
			public static class Domain {
				private int from;
				private int to;
				private double confidence;
				private String value;
				
				public void setFrom(int from) {
					this.from = from;
				}
				public int getFrom() {
					return from;
				}
				public void setTo(int to) {
					this.to = to;
				}
				public int getTo() {
					return to;
				}
				public void setConfidence(double confidence) {
					this.confidence = confidence;
				}
				public double getConfidence() {
					return confidence;
				}
				public void setValue(String value) {
					this.value = value;
				}
				public String getValue() {
					return value;
				}
			}
			
			private int length;
			private List<Domain> domains;
			
			public int getLength() {
				return length;
			}
			
			public void setLength(int length) {
				this.length = length;
			}
			
			public List<Domain> getDomains() {
				if (domains == null)
					domains = new ArrayList<Domain>();
				return domains;
			}
		}
		private String symbol;
		private String accession;
		private String name;
		private String location;
		private String uri;
		private DomainArchitecture domainArchitecture;
		private List<Annotation> annotations;
		
		public void setSymbol(String symbol) {
			this.symbol = symbol;
		}
		public String getSymbol() {
			return symbol;
		}
		public void setAccession(String accession) {
			this.accession = accession;
		}
		public String getAccession() {
			return accession;
		}
		public void setName(String name) {
			this.name = name;
		}
		public String getName() {
			return name;
		}
		public void setLocation(String location) {
			this.location = location;
		}
		public String getLocation() {
			return location;
		}
		public void setUri(String uri) {
			this.uri = uri;
		}
		public String getUri() {
			return uri;
		}	
		public DomainArchitecture getDomainArchitecture() {
			return domainArchitecture;
		}
		public void setDomainArchitecture(DomainArchitecture domainArchitecture) {
			this.domainArchitecture = domainArchitecture;
		}
		public List<Annotation> getAnnotations() {
			if (annotations == null)
				annotations = new ArrayList<Annotation>();
			return annotations;
		}
	}
	
	public static class Reference {
		private String doi;

		public void setDoi(String doi) {
			this.doi = doi;
		}

		public String getDoi() {
			return doi;
		}
	}
	
	private static final long serialVersionUID = 1L;
	private double branchLength;
    private Taxonomy taxonomy;
    private List<Distribution> distributions;
    private Sequence sequence;
    private Color color;
    private List<Confidence> confidences;
    private List<Property> properties;
    private List<Reference> references;
    private int width;

    /**
     * Default constructor. Null label, branch length equals one.
     */
    public Clade() {
        this(null, 1, 0);
    }

    /**
     * Constructor. 
     * @param nodeLabel optional label for the node
     * @param branchLength non-negative distance
     */
    public Clade(String nodeLabel, double branchLength) {
        this(nodeLabel, branchLength, 0);
    }

    /**
     * Constructor.
     * @param nodeLabel optional label for the node
     * @param branchLength non-negative distance
     * @param confidence non-negative confidence value
     */
    public Clade(String nodeLabel, double branchLength, double confidence) {
        super(nodeLabel);
        this.branchLength = branchLength;
        
        if (confidence != 0) {
	        Confidence conf = new Confidence();
	        conf.setType("unknown");
	        conf.setValue(confidence);
	        getConfidences().add(conf);
        }
    }

	/**
     * Branch length.
     * @return non-negative distance
     */
    public double getBranchLength() {
        return branchLength;
    }

    /**
     * Bootstrap replicates, or zero (0) if there aren't any.
     * @return number of replicates; non-negative
     */
    public double getConfidence() {
    	Confidence conf = confidences == null ? null : confidences.get(0);
        return conf == null ? 0 : conf.getValue();
    }

    /**
     * Branch length.
     * @param distance non-negative length
     */
    public void setBranchLength(double distance) {
    	if(distance < 0)
    		throw new IllegalArgumentException("Branch length cannot be negative.");
        branchLength = distance;
    }

    /**
     * Bootstrap replicates, e.g., or zero (0) if there aren't any.
     * @param replicates non-negative value
     */
    public void setConfidence(double replicates) {
    	if(replicates < 0)
    		throw new IllegalArgumentException("Bootstrap replicates cannot be negative.");
    	Confidence conf = new Confidence();
    	conf.setType("unknown");
    	conf.setValue(replicates);
        confidences = Collections.singletonList(conf);
    }

    /**
     * Overridden to display bootstrap replicates for non-leaf nodes.
     */
    public String toString() {    	
    	String s = super.toString();
        if (s == null && taxonomy != null) {
        	s = taxonomy.code == null 
        		? taxonomy.scientificName == null 
        			? taxonomy.commonName == null
        				? null
        				: taxonomy.commonName
        			: taxonomy.scientificName
        		: taxonomy.code;
        }
        if (s == null && distributions != null && !distributions.isEmpty()) {
        	s = "";
        	for (Distribution d : distributions) {
        		if (d.getDescription() != null) {
        			if (s.length() > 0)
        				s += ", ";
        			s += d.getDescription();
        		}
        	}
        }
        if (s == null && getConfidence() != 0) {
        	s = String.valueOf(getConfidence());
        }        
        return s;
    }

	public Taxonomy getTaxonomy() {
		return taxonomy;
	}
	
	public void setTaxonomy(Taxonomy taxonomy) {
		this.taxonomy = taxonomy;
	}
	
	public void setColor(Color color) {
		this.color = color;
	}

	public Color getColor() {
		return color;
	}

	public void setWidth(int width) {
		this.width = width;
	}

	public int getWidth() {
		return width;
	}

	public List<Distribution> getDistributions() {
		if (distributions == null)
			distributions = new ArrayList<Distribution>();
		return distributions;
	}

	public String getToolTipText() {
		String tt = "<html>";
		String indent = "<div style=\"padding-left:8px\">";
		
		boolean hasName = getUserObject() != null;
		
		if (hasName) {
			tt += "<div style=\"font-size:10px;font-weight:bold\">" 
				+ getUserObject() + "</div>";
		}
		
		if (confidences != null && !confidences.isEmpty()) {
			tt += "<div style=\"font-weight:bold\">Confidence:</div>";
			for (Confidence conf : confidences) {
				tt += indent + conf.getType() + ": " + conf.getValue() + "</div>";
			}
		}
			
		if (taxonomy != null) {
			tt += "<div style=\"font-weight:bold\">Taxonomy:</div>";
			if (taxonomy.code != null) {
				tt += indent + "Code: " + taxonomy.code + "</div>";
			}
			if (taxonomy.scientificName != null) {
				tt += indent + "Scientific Name: " + taxonomy.scientificName + "</div>";
			}
			if (taxonomy.commonName != null) {
				tt += indent + "Common Name: " + taxonomy.commonName + "</div>";
			}
			if (taxonomy.rank != null) {
				tt += indent + "Rank: " + taxonomy.rank + "</div>";
			}
			if (taxonomy.uri != null) {
				//tt += indent + "Uri: " + taxonomy.uri + "</div>";
			}
		}
		
		if (sequence != null) {
			tt += "<div style=\"font-weight:bold\">Sequence:</div>";
			if (sequence.symbol != null) {
				tt += indent + "Symbol: " + sequence.symbol + "</div>";
			}
			if (sequence.accession != null) {
				tt += indent + "Accession: " + sequence.accession + "</div>";
			}
			if (sequence.name != null) {
				tt += indent + "Name: " + sequence.name + "</div>";
			}
			if (sequence.location != null) {
				tt += indent + "Location: " + sequence.location + "</div>";
			}
			if (sequence.uri != null) {
				//tt += indent + "Uri: " + sequence.uri + "</div>";
			}
			if (sequence.annotations != null && !sequence.annotations.isEmpty()) {
				tt += indent + "Annotations:";
				for (Annotation a : sequence.annotations) {
					if (a.description != null) {
						tt += indent + "Description: " + a.description + "</div>";
					}
				}
				tt += "</div>";
			}			
			if (sequence.domainArchitecture != null) {
				tt += indent + "Length: " + sequence.domainArchitecture.length + "</div>";
				if (sequence.domainArchitecture.domains != null) {
					tt += indent + "Domains: ";
					String s = "";
					for (int i = 0; 
							i < sequence.domainArchitecture.domains.size(); 
							++i) {
						Domain d = sequence.domainArchitecture.domains.get(i);
						if (i % 6 > 0)
							s += ", ";
						s += d.value;
						if ((i + 1) % 6 == 0 
								&& i < sequence.domainArchitecture.domains.size() - 1) 
							s += ",</div><div style=\"padding-left:16px\">";
					}
					tt += s + "</div>";
				}
			}
		}
		
		if (distributions != null && !distributions.isEmpty()) {
			boolean hasDesc = false;
			for (Distribution dist : distributions) {
				if (dist.getDescription() != null) {
					if (!hasDesc) {
						tt += "<div style=\"font-weight:bold\">Distribution:</div>";
						hasDesc = true;
					}
					tt += indent + dist.getDescription() + "</div>";
				}
			}			
		}
		
		tt += "</html>";
		return tt;
	}

	public Sequence getSequence() {
		return sequence;
	}
	
	public void setSequence(Sequence sequence) {
		this.sequence = sequence;
	}
	
	public List<Confidence> getConfidences() {
		if (confidences == null)
			confidences = new ArrayList<Confidence>();
		return confidences;
	}

	public List<Reference> getReferences() {
		if (references == null)
			references = new ArrayList<Reference>();
		return references;
	}
	
	public List<Property> getProperties() {
		if (properties == null)
			properties = new ArrayList<Property>();
		return properties;
	}
}
